! class methods for StandFields, RtBrSort
  MEMBER
_ABCDllMode_  EQUATE(0)
_ABCLinkMode_ EQUATE(1)

! Class to allow user to append sort order from selected fields at runtime.
! if used on tabs without sort order the sort order is set to the one created.
! if used on tabs with a sort order the new sort is appended to the current one.
! on tabs with design time sort, the fields will from the key will not be avaliable

! updates, see header in inc file

  ! local procedures to be used
  map  
    GetSortFields( StandFieldQueue Flds, RtSortTabs SCtrls, byte KeyNumber, *byte LocateFree ),byte
    CaseField( *string Fld, byte UpCase ),string
  .
  include('AbWindow.inc')
  include('keycodes.clw')
  include('RtBrSort.inc')

! ---------------------------------------------------------------------------------------
!  Begin class methods
!
!  StandFields
! ---------------------------------------------------------------------------------------
StandFields.Init         procedure
! init method, create new one 

  code

  if (Self.Fields &= NULL)          ! if not there then new one
    Self.Fields &= new(StandFieldQueue)
  end ! if

  return
! StandFields.Init --------------------------------------------------------------------------

StandFields.Kill         procedure
! dispose the queue

x                        byte,auto

  code

  ! step thru queue and null the any variable
  loop x = 1 to records(Self.Fields)
    get(Self.Fields, x)
    Self.Fields.FieldActual &= NULL
  end
  ! dump the rest
  dispose(Self.Fields)

  return
! StandFields.Kill ----------------------------------------------------------------------------

StandFields.AddField    procedure( *? FldActual, string FldDisplay, string FldName, byte KNumber, byte UpCase, string FldPicture, byte FldType, byte FldNumber )
! add fields to the queue 

  code

  ASSERT(~(SELF.Fields &= NULL))            ! make sure it is there
  clear(Self.Fields)
  Self.Fields.FieldActual &= FldActual      ! assign the any for the locator
  Self.Fields.FieldDisplay = FldDisplay
  Self.Fields.FieldName = FldName
  Self.Fields.KeyNumber = KNumber
  Self.Fields.CaseSensitive = UpCase
  Self.Fields.FieldPicture = FldPicture
  Self.Fields.FieldType = FldType
  Self.Fields.FieldNumber = FldNumber
  add(Self.Fields)
  assert(~errorcode())                      ! if the add failed throw it

  return
! StandFields.AddFields -----------------------------------------------------------------------
!
! End StandFields methods
!
! Begin RtBrSOrt methods
! ---------------------------------------------------------------------------------------------

RtBrSort.AddTabSort   procedure( byte CtrlNum, string Txt, byte UseText )
! proc to add a tb to the queue

  code

  ASSERT(~(SELF.Sorts &= NULL))        ! make sure it is there
  Self.Sorts.TabControl = CtrlNum      ! assign passed info
  Self.Sorts.TabText = Txt
  Self.Sorts.UseTabText = UseText
  Self.Sorts.SortStr = ''              ! make sure both strings are empty
  Self.Sorts.SortDisplay = ''
  add(Self.Sorts)
  assert(~errorcode())                 ! if the add failed throw it

  return
! RtBrSort.AddTabSort --------------------------------------------------------------------------

RtBrSort.ClearSort   procedure()
! procedure to clear the sort and return the order property to the original
! order clause, calls AdjustSort to set order

  code

  ! clear the strings and put the queue record
  Self.Sorts.SortStr = ''
  Self.Sorts.SortDisplay = ''
  put(Self.Sorts)            ! if the put failed throw it
  assert(~errorcode())
  if ~Self.Sort.Locator &= NULL     ! if locator present then clear and reset
    Self.Sort.Locator.Destruct()
    Self.Sort.Locator.Reset()       ! hide/unhide the locator
  end !if
  Self.AdjustSort(true, Self.ToTopReset, 0)
  

  return
! RtBrSort.ClearSort --------------------------------------------------------------------------

RtBrSort.AdjustSort   procedure( byte Force, byte ToTop, byte LocateFree )

  code

  ! the KeyToOrder method (FileManager) will build sort order clause based on the keys
  ! of the current sort order, if no key blank is returned
  Self.SetOrder(Self.Primary.me.KeyToOrder(Self.Order.MainKey, 1))
  ! if some thing there, ie user did not press clear then okay (string will be blank)
  ! append sort string to current sort order
  if (Self.Sorts.SortStr)
    Self.AppendOrder(Self.Sorts.SortStr)
  end
  ! reset the sort order and update tabs
  parent.ResetSort(Force)
  if (ToTop)
    ! scroll to top if user selected
    Self.TakeScroll(Event:ScrollTop)
  end !if
  if LocateFree                       ! if > 0 then user selected a field
    get(Self.RtFields.Fields, LocateFree)
    if ~Self.Sort.Locator &= NULL     ! is there a locator
      Self.Sort.Locator.Destruct()            ! null the any
      ! assign the new value to the free element
      Self.sort.Locator.FreeElement &= Self.RtFields.Fields.FieldActual
      ! set input stuff
      case Self.RtFields.Fields.FieldType
        of Fte:UpperString
          Self.Sort.Locator.Control{prop:upr} = true
        of Fte:CapString
            Self.Sort.Locator.Control{prop:cap} = true
      end !case
      ! and set picture
      if Self.Sort.Locator.Control
        Self.Sort.Locator.Control{prop:text} = Self.RtFields.Fields.FieldPicture
      end
    end !if ~Self.Sort.Locator
  end !if LocateFree
  Self.UpdateTabs()

  return
! RtBrSOrt.AdjustSort ----------------------------------------------------------------------

RtBrSort.GetNewSort  procedure()
! proc to change the current sort order

LocateFree     byte,auto

  code

  ! call local proc to get new fields from the user for the tab selected
  if GetSortFields( Self.RtFields.Fields, Self.Sorts, choice(Self.SheetControl), LocateFree)
    put(Self.Sorts)            ! if the put failed throw it
    assert(~errorcode())
    Self.AdjustSort(true, Self.ToTopAppend, LocateFree)
    Self.Sort.Locator.Reset()        ! hide/unhide the locator
  end !if

  return
! RtBrSort.GetNewSort -------------------------------------------------------------------------

RtBrSort.Init   procedure(signed ListBox,*string Posit,view V,queue Q,RelationManager RM,WindowManager WM)
! call the init method for the parent, BrowseClass and then call the method
! init the fields queue 

 code

  parent.Init(ListBox, Posit, v, q, Rm, Wm)
  Self.InitFields()

  return

! RtBrSort.Init -------------------------------------------------------------------------------

RtBrSort.InitControls  procedure(byte SortNum, byte ClrNum, byte ShtNum, byte StrCtrl, long SortKey = AltA, long ClearKey = AltR )
! init RtBrSort class controls and keycodes

  code

  Self.SortControl = SortNum
  Self.ResetControl = ClrNum
  Self.SheetControl = ShtNum
  Self.StringControl = StrCtrl
  ! assign new key codes to properties
  Self.SetKeyCodes( SortKey, ClearKey )

  return
! RtBrSort.Init -------------------------------------------------------------------------------

RtBrSort.InitFields   procedure()
! init the fields for the class

  code

  if (Self.RtFields &= NULL)            ! if not there make a new one
    Self.RtFields &= new(StandFields)
    Self.RtFields.Init()                ! make new fields
  end !if
  if (Self.Sorts &= NULL)               ! make a new sort tab queue
    Self.Sorts &= new(RtSortTabs)
  end

  return
! RtBrSort.InitFields -------------------------------------------------------------------------

RtBrSort.Kill      procedure()
! get ride of memory used,
! and call kill method for browse class

  code

  Self.RtFields.Kill()      ! kill the fields reference
  dispose(Self.RtFields)    ! dispose the class refrence
  dispose(Self.Sorts)       ! dispose the queue
  parent.Kill()             ! browse class

  return
! RtBrSort.Kill ------------------------------------------------------------------------------

RtBrSort.SetScrollAction     procedure( byte TopA = true, byte TopR = false )
! set the scroll action for changes in sort order
! Note the default values

  code

  Self.ToTopAppend = TopA
  Self.ToTopReset = TopR

  return
! RtBrSort.SetScrollAction --------------------------------------------------------------------

RtBrSort.SetKeyCodes  procedure( long AppendKey, long ResetKey )
! method to change the key codes used by the RT sort class
! defaults are AltA and AltR

  code

  ! assign new key codes to properties
  Self.AppendKeyCode = AppendKey
  Self.ListControl{Prop:Alrt,255} = AppendKey
  Self.ResetKeyCode = ResetKey
  Self.ListControl{Prop:Alrt,255} = ResetKey

  return
! RtBrSort.ChangeKeyCodes ---------------------------------------------------------------------

RtBrSort.SetSort   procedure(byte NewOrder, byte Force)  ! byte,virtual,proc
! derived method, called when user selects a new tab or sort order
! find the tab selected and set the sort to active

ReturnValue        byte,auto
x                  byte,auto        ! local counter

  code

  ReturnValue = Parent.SetSort(NewOrder, Force)
  if ReturnValue
    Self.Active = false                   ! set flag
    loop x = 1 to records(Self.Sorts)     ! step thru queue to find
      get(Self.Sorts, x)
      assert(~errorcode())                 ! if there is an error here we got major trouble
      ! if the feq from the tab control = the class property then we have a active RT sort
      ! set flag and end loop,
      if Self.SheetControl{Prop:ChoiceFeq} = Self.Sorts.TabControl
        Self.Active = true
        break
      end
    end !loop
  end !if ReturnValue
  
  return(ReturnValue)
! RtBrSort.SetSort -----------------------------------------------------------------------------

RtBrSort.TakeEvent procedure
! derived method
! parent called last, because method is not a function
! only two events needed in this method, one of the RT Sort controls
! if one accepted then call needed method if not then let parent deal with it

  code

  CASE ACCEPTED()               ! which control was accepted
    of SELF.SortControl         ! one of class' then call method
      Self.GetNewSort
      select(Self.ListControl)
    of Self.ResetControl
      Self.ClearSort()
      select(Self.ListControl)
  else
    parent.TakeEvent()          ! let parent handle
  end ! case

  return
! RtBrSort.TakeEvent ---------------------------------------------------------------------------

RtBrSort.TakeKey   procedure()
! derived method,
! handle the two key codes for the class, 

ReturnValue        byte(false)       ! set by method, true if key was accepted

  code

  ! test for RT Sort key codes
  ! if there are records, and this tab is active
  if records(SELF.ListQueue) and Self.Active 
    case keycode() 
      of Self.AppendKeyCode           ! if this key then post event
        if Self.SortControl           ! if control present
          post(Event:Accepted, Self.SortControl)
          ReturnValue = true
        end !if
      of Self.ResetKeyCode            ! if this key then post event
        if Self.ResetControl          ! if control pesent
          post(Event:Accepted, Self.ResetControl)
          ReturnValue = true
        end !if
    end ! case keycode
  end !if
  ! call parent if key was not handled here
  if ~ReturnValue
    ReturnValue = Parent.TakeKey()
  end

  return(ReturnValue)

! RtBrSort.TakeKey ------------------------------------------------------------------------

RtBrSort.UpdateTabs     procedure()
! method to update the tab display,
! if there is a sort string then change the tab text messages
! if no sort string then use default text

  code

  if Self.Sorts.SortStr           ! sort present
    if Self.StringControl         ! if string control in use then message goes there
      Self.StringControl{prop:use} = Self.Sorts.SortDisplay
    else                          ! if not then use the tabs
      if Self.Sorts.UseTabText    ! use the default with RT display
        Self.Sorts.TabControl{prop:text} = clip(Self.Sorts.TabText) & ',' & Self.Sorts.SortDisplay
      else                        ! do not use the default
        Self.Sorts.TabControl{prop:text} = Self.Sorts.SortDisplay
      end ! if
    end
  else                            ! use default, after a clear sort
    Self.Sorts.TabControl{prop:text} = Self.Sorts.TabText
    Self.StringControl{prop:use} = ''
  end !if

  return
! RtBrSort.UpdateTabs ------------------------------------------------------------------------

RtBrSort.UpdateWindow   procedure()
! derived method,
! update the controls on the screen

  code

  if records(Self.ListQueue) and Self.Active         ! enable control if both true
    if Self.SortControl
      enable(Self.SortControl)
    END
    if Self.ResetControl and Self.Sorts.SortStr ! if sort str present need clear
      enable(Self.ResetControl)
    else
      disable(Self.ResetControl)                ! nothing to clear yet
    END
 else                                       ! disable if either are false
   if Self.SortControl
      disable(Self.SortControl)
    END
    if Self.ResetControl
      disable(Self.ResetControl)
    END
 end
 ! call the parent to update other controls
 parent.UpdateWindow()

 return
! RtBrSort.UpdateWindow -------------------------------------------------------------------

CaseField   procedure( *string Fld, byte CaseSensitive )

RetStr      string(150)

  code

  if ~CaseSensitive
    RetStr = 'UPPER(' & clip(Fld) & ')'
  else
    RetStr = Fld
  end

  return(RetStr)

! CaseField ------------------------------------------------------------------------------

! procedure to let the user select from list of fields to sort on
! proc is local to the class module will not be called from any other procedures
! so might as well be here.
GetSortFields   procedure( StandFieldQueue WorkFlds, RtSortTabs SCtrls, byte KeyNumber, *byte LocateFree)

! -------------------- Begin window input proc --------------------------------------------

LocalResponse   long,auto               ! dummy
WorkSort             STRING(150)        ! string to hold input
WorkDisp             STRING(100)        ! string to display input
AorD                 STRING('+')        ! local value for asc or des sorts
x                    byte,auto          ! local counter
WorkLocate           byte,auto
! -------------------------------------------------------------------------------------------

window WINDOW('Select Sort Order'),AT(,,227,94),FONT('MS Sans Serif',8,,FONT:regular),CENTER,IMM,  |
         SYSTEM,GRAY
       BUTTON('&Ascending'),AT(6,8,47,13),USE(?AscButton)
       BUTTON('&Descending'),AT(56,8,46,13),USE(?DesButton)
       BUTTON('C&lear'),AT(104,8,46,13),USE(?ClearButton)
       LIST,AT(7,31,89,10),USE(?FieldList),vscroll,DISABLE,FORMAT('100L(1)@S25@'),DROP(10),FROM(WorkFlds)
       ENTRY(@s100),AT(7,52,210,10),USE(WorkDisp),SKIP,DISABLE,READONLY
       BUTTON('&OK'),AT(45,72,35,14),USE(?OkButton),DEFAULT
       BUTTON('&Cancel'),AT(145,72,36,14),USE(?CancelButton)
     END

ThisWindow           CLASS(WindowManager)
Init                   PROCEDURE(),BYTE,PROC,VIRTUAL
TakeAccepted           PROCEDURE(),BYTE,PROC,VIRTUAL
Kill                   PROCEDURE(),BYTE,PROC,VIRTUAL
                     END
! -----------------------------------------------------------------------------------
! this is standard window manager 

  CODE

  LocalResponse = ThisWindow.Run()
  if ThisWindow.Response = RequestCancelled
    return(0)
  else
    return(ThisWindow.Response)
  end ! if
! -----------------------------------------------------------------------------------

ThisWindow.Init PROCEDURE()

ReturnValue          BYTE,AUTO

  CODE

  ! assign current to workers
  WorkDisp = SCtrls.SortDisplay
  WorkSort = SCtrls.SortStr

  ReturnValue = PARENT.Init()
  IF ReturnValue THEN RETURN ReturnValue.
  SELF.FirstField = ?AscButton
  OPEN(window)
  SELF.Opened=True
  SELF.SetAlerts()

  RETURN ReturnValue

! Init ------------------------------------------------------------------------------------

ThisWindow.Kill PROCEDURE()

ReturnValue          BYTE,AUTO

  CODE

  ReturnValue = PARENT.Kill()
  return(ReturnValue)
! -------------------------------------------------------------------------------------------
ThisWindow.TakeAccepted PROCEDURE()

ReturnValue          BYTE,AUTO
Looped BYTE
  CODE
  LOOP
    IF Looped
      RETURN Level:Notify
    ELSE
      Looped = 1
    END
    CASE ACCEPTED()
    ! set variable to button pressed
    OF ?AscButton
      Aord = '+'
      do Dis_EnButtons
    OF ?DesButton
      AorD = '-'
      do Dis_EnButtons
    OF ?FieldList
      ! get the queue, based on the user choice
      get(WorkFlds, choice(?FieldList))
      if WorkSort
        ! if there is something there then add new to the end
        WorkSort = clip(WorkSort) & ',' & AorD & CaseField(WorkFlds.FieldName, WorkFlds.CaseSensitive)
        WorkDisp = clip(WorkDisp) & ',' & AorD & WorkFlds.FieldDisplay
      else
        WorkLocate = WorkFlds.FieldNumber
        WorkSort = AorD & CaseField(WorkFlds.FieldName, WorkFlds.CaseSensitive)
        WorkDisp = AorD & WorkFlds.FieldDisplay
      end
      do Dis_EnButtons
      display(?WorkDisp)

    of ?ClearButton
      WorkSort = ''
      WorkDisp = ''
      WorkLocate = 0
      display(?WorkDisp)

    OF ?OkButton
      ! user is happy so fill the para passed 
      SCtrls.SortStr = WorkSort
      SCtrls.SortDisplay = WorkDisp
      LocateFree = WorkLocate
      Self.Response = RequestCompleted
      post(Event:CloseWindow)

    OF ?CancelButton
      ! no changes so get out
      Self.Response = RequestCancelled
      post(Event:CloseWindow)
    END ! case

    ReturnValue = PARENT.TakeAccepted()
    RETURN ReturnValue
  END

  ReturnValue = Level:Fatal

  RETURN ReturnValue
! end procedure body ---------------------------------------------------------

! begin procedure Rtns
! ----------------------------------------------------------------------------
Dis_EnButtons       routine
! rtn to enable/disable the buttons and the drop list, user has to add
! in sequence

  if (?AscButton{prop:disable})
    enable(?AscButton, ?ClearButton)
    disable(?FieldList)
  else
    disable(?AscButton, ?ClearButton)
    enable(?FieldList)
  end

  exit
! end procedure Rtns
! -------------------- End window input proc --------------------------------------------
!
! begin locator class methods
! ----------------------------------------------------------------------------------------

NoKeyEntryLocator.Init   PROCEDURE(SIGNED Control = 0, SIGNED FldPrompt = 0, *? Free,BYTE NoCase = 0, <BrowseClass B>)
! init method for Entry locator, call parent to init and
! set free element to NULL, Free passed is a place holder for the init method

  code

  Parent.Init( Control, Free, NoCase, B )
  Self.Destruct()                           ! null at start, no sort active
  Self.FieldPrompt = FldPrompt              ! assign prompt

  return
! NoKeyLocator.Init-----------------------------------------------------------------------

NoKeyEntryLocator.ReSet    procedure
! use the LocatorClass virtual method Reset to hide or unhide the locator.

  code

  if ~Self.FreeElement &= NULL
    Self.SetHide(false)
  else
    Self.SetHide(true)
  end ! if ~Self.FreeElement

  if Self.Control   ! set both properties to false for new free element
    Self.Control{prop:upr} = false
    Self.Control{prop:cap} = false
  end ! if Self.Control

  return
! NoKeyLocator.SetFree -------------------------------------------------------------------

NoKeyEntryLocator.TakeAccepted  PROCEDURE !VIRTUAL,BYTE 1 for locator on value required
! method to locate value

  code

  update(Self.Control)
  ! fill the free element
  Self.FreeElement = Self.Shadow
  if Self.FreeElement               ! something there then reset and select list
    Self.ViewManager.ResetFromBuffer()
    select(Self.ViewManager.ListControl)
  end ! if

  return(false)         ! always false
! NoKeyLocator.TakeAccepted --------------------------------------------------------------

NoKeyEntryLocator.SetHide   procedure( byte h )
! method to hide or unhide the locator control and prompt
! control is hidden when no sort order is active, avoids user problems

  code

  if Self.Control                       ! if the is a control
    Self.Control{prop:hide} = h         ! set property
    if Self.FieldPrompt                 ! nested no control no prompt
      Self.FieldPrompt{prop:hide} = h
    end !if
  end !if Self.Control

  return
! NoKeyEnteryLocator.SetHide --------------------------------------------------------------

NoKeyIncrementLocator.TakeKey   PROCEDURE
! incremental locator TakeKey, mainly the same as the standard locator
! but calls the reset from buffer method, and always return false

Key           byte,auto
ChangeBuffer  byte(false)       ! local flag

  code

  if ~Self.FreeElement &= NULL      ! is sort active
    Key = keychar()
    if keychar() = BSKey
      if len(SELF.Shadow)
        Self.Shadow = Self.Shadow[1:len(Self.Shadow) - 1]
      END
      ChangeBuffer = true       ! set flag
    elsif keycode() = SpaceKey OR Key
      Self.Shadow = SELF.Shadow & choose(keycode()=SpaceKey,' ',chr(Key))
      ChangeBuffer = true       ! set flag
    END

    if ChangeBuffer             ! something there so reset
      Self.Update
      Self.ViewManager.ResetFromBuffer()
    end !if ChangeBuffer
  end  !if Self.Control

  return(false)     ! always false
! NoKeyIncrementLocator.TakeKey --------------------------------------------------

NoKeyStepLocator.Init   PROCEDURE(SIGNED Control = 0, *? Free,BYTE NoCase = 0, <BrowseClass B>)
! init method for Step locator, call parent to init and
! set free element to NULL, Free passed is a place holder for the init method

  code

  parent.init(Control, Free, NoCase, B)
  Self.FreeElement &= NULL              ! no need until sort order active

  return
! NoKeyStepLocator.Init -----------------------------------------------------------

NoKeyStepLocator.TakeKey  procedure ! ,VIRTUAL,BYTE ! 1 for locator on value required

  code

  if ~Self.FreeElement &= NULL      ! is sort active
    if parent.TakeKey()             ! if parent accepted key then reset
      Self.ViewManager.ResetFromBuffer()
    end ! if parent
  end

  return(false)     ! always return false
! -------------------------------------------------------------